def synchronized_with_attr(attr_name):

    def decorator(func):

        def synced_func(*args, **kws):
            self = args[0]
            lock = getattr(self, attr_name)
            with lock:
                return func(*args, **kws)
        return synced_func

    return decorator


def truncate(ori_str, length=100):
    if not ori_str:
        return ""
    return ori_str[:length] + "..." if len(ori_str) > length else ori_str


import os.path
import fcntl
import logging
import sys

logger = logging.getLogger("acm")


def read_file(base, key):
    file_path = os.path.join(base, key)
    if not os.path.exists(file_path):
        return None

    try:
        if sys.version_info[0] == 3:
            with open(file_path, "r+", newline="") as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                return f.read()
        else:
            with open(file_path, "r+") as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                return f.read()
    except OSError:
        logger.exception("[read-file] read file failed, file path:%s" % file_path)
        return None


def save_file(base, key, content):
    file_path = os.path.join(base, key)
    if not os.path.isdir(base):
        try:
            os.makedirs(base)
        except OSError:
            logger.warning("[save-file] dir %s is already exist" % base)

    try:
        with open(file_path, "wb") as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            f.write(content if type(content) == bytes else content.encode("utf8"))

    except OSError:
        logger.exception("[save-file] save file failed, file path:%s" % file_path)


def delete_file(base, key):
    file_path = os.path.join(base, key)
    try:
        os.remove(file_path)
    except OSError:
        logger.warning("[delete-file] file not exists, file path:%s" % file_path)



from setuptools import setup, find_packages
import acm


long_description = open('README.md').read()

setup(
    name="acm-sdk-python",
    version=acm.__version__,
    packages=find_packages(exclude=["test"]),
    url="https://github.com/alibaba/acm-sdk-python",
    license="Apache License 2.0",
    author="acm",
    author_email="755063194@qq.com",
    description="Python client for ACM.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[],
    entry_points={
      "console_scripts": ["acm=acm.command:main"],
    },
)



import base64
import hashlib
import hmac
import logging
import socket
import time
import ssl
import sys
import json

from multiprocessing import Process, Manager, Queue, pool
from threading import RLock, Thread

try:
    # python3.6
    from http import HTTPStatus
    from urllib.request import Request, urlopen
    from urllib.parse import urlencode, unquote_plus
    from urllib.error import HTTPError, URLError
except ImportError:
    # python2.7
    import httplib as HTTPStatus
    from urllib2 import Request, urlopen, HTTPError, URLError
    from urllib import urlencode, unquote_plus

    base64.encodebytes = base64.encodestring

from .commons import synchronized_with_attr, truncate
from .params import group_key, parse_key, is_valid
from .server import get_server_list
from .files import read_file, save_file, delete_file

logging.basicConfig()
logger = logging.getLogger()

DEBUG = False
VERSION = "0.3.10"

DEFAULT_GROUP_NAME = "DEFAULT_GROUP"
DEFAULT_NAMESPACE = ""

WORD_SEPARATOR = u'\x02'
LINE_SEPARATOR = u'\x01'

kms_available = False

try:
    from aliyunsdkcore.client import AcsClient
    from aliyunsdkkms.request.v20160120.DecryptRequest import DecryptRequest
    from aliyunsdkkms.request.v20160120.EncryptRequest import EncryptRequest

    kms_available = True
except ImportError:
    logger.info("Aliyun KMS SDK is not installed")

ENCRYPTED_DATA_ID_PREFIX = "cipher-"

DEFAULTS = {
    "APP_NAME": "ACM-SDK-Python",
    "TIMEOUT": 3,  # in seconds
    "PULLING_TIMEOUT": 30,  # in seconds
    "PULLING_CONFIG_SIZE": 3000,
    "CALLBACK_THREAD_NUM": 10,
    "FAILOVER_BASE": "acm-data/data",
    "SNAPSHOT_BASE": "acm-data/snapshot",
    "KMS_ENABLED": False,
    "REGION_ID": "",
    "KEY_ID": "",
}

OPTIONS = set(
    ["default_timeout", "tls_enabled", "auth_enabled", "cai_enabled", "pulling_timeout", "pulling_config_size",
     "callback_thread_num", "failover_base", "snapshot_base", "app_name", "kms_enabled", "region_id",
     "kms_ak", "kms_secret", "key_id"])


class ACMException(Exception):
    pass


class ACMRequestException(ACMException):
    pass


def process_common_params(data_id, group):
    if not group or not group.strip():
        group = DEFAULT_GROUP_NAME
    else:
        group = group.strip()

    if not data_id or not is_valid(data_id):
        raise ACMException("Invalid dataId.")

    if not is_valid(group):
        raise ACMException("Invalid group.")
    return data_id, group


def parse_pulling_result(result):
    if not result:
        return list()
    ret = list()
    for i in unquote_plus(result.decode()).split(LINE_SEPARATOR):
        if not i.strip():
            continue
        sp = i.split(WORD_SEPARATOR)
        if len(sp) < 3:
            sp.append("")
        ret.append(sp)
    return ret


def is_encrypted(data_id):
    return data_id.startswith(ENCRYPTED_DATA_ID_PREFIX)


class WatcherWrap:
    def __init__(self, key, callback):
        self.callback = callback
        self.last_md5 = None
        self.watch_key = key


class CacheData:
    def __init__(self, key, client):
        self.key = key
        local_value = read_file(client.failover_base, key) or read_file(client.snapshot_base, key)
        self.content = local_value
        src = local_value.decode("utf8") if type(local_value) == bytes else local_value
        self.md5 = hashlib.md5(src.encode("GBK")).hexdigest() if src else None
        self.is_init = True
        if not self.md5:
            logger.debug("[init-cache] cache for %s does not have local value" % key)


class ACMClient:
    """Client for ACM

    available API:
    * get
    * add_watcher
    * remove_watcher
    """
    debug = False

    @staticmethod
    def set_debugging():
        if not ACMClient.debug:
            global logger
            logger = logging.getLogger("acm")
            handler = logging.StreamHandler()
            handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s:%(message)s"))
            logger.addHandler(handler)
            logger.setLevel(logging.DEBUG)
            ACMClient.debug = True

    def __init__(self, endpoint, namespace=None, ak=None, sk=None, ):
        self.endpoint = endpoint
        self.namespace = namespace or DEFAULT_NAMESPACE or ""
        self.ak = ak
        self.sk = sk

        self.server_list = None
        self.server_list_lock = RLock()
        self.current_server = None
        self.server_offset = 0
        self.server_refresh_running = False

        self.watcher_mapping = dict()
        self.pulling_lock = RLock()
        self.puller_mapping = None
        self.notify_queue = None
        self.callback_tread_pool = None
        self.process_mgr = None

        self.default_timeout = DEFAULTS["TIMEOUT"]
        self.tls_enabled = False
        self.auth_enabled = self.ak and self.sk
        self.cai_enabled = True
        self.pulling_timeout = DEFAULTS["PULLING_TIMEOUT"]
        self.pulling_config_size = DEFAULTS["PULLING_CONFIG_SIZE"]
        self.callback_tread_num = DEFAULTS["CALLBACK_THREAD_NUM"]
        self.failover_base = DEFAULTS["FAILOVER_BASE"]
        self.snapshot_base = DEFAULTS["SNAPSHOT_BASE"]
        self.app_name = DEFAULTS["APP_NAME"]
        self.kms_enabled = DEFAULTS["KMS_ENABLED"]
        self.region_id = DEFAULTS["REGION_ID"]
        self.key_id = DEFAULTS["KEY_ID"]
        self.kms_ak = self.ak
        self.kms_secret = self.sk
        self.kms_client = None

        logger.info("[client-init] endpoint:%s, tenant:%s" % (endpoint, namespace))

    def set_options(self, **kwargs):
        for k, v in kwargs.items():
            if k not in OPTIONS:
                logger.warning("[set_options] unknown option:%s, ignored" % k)
                continue

            if k == "kms_enabled" and v and not kms_available:
                logger.warning("[set_options] kms can not be turned on with no KMS SDK installed")
                continue

            logger.debug("[set_options] key:%s, value:%s" % (k, v))
            setattr(self, k, v)

    def _refresh_server_list(self):
        with self.server_list_lock:
            if self.server_refresh_running:
                logger.warning("[refresh-server] task is running, aborting")
                return
            self.server_refresh_running = True

        while True:
            try:
                time.sleep(30)
                logger.debug("[refresh-server] try to refresh server list")
                server_list = get_server_list(self.endpoint, 443 if self.tls_enabled else 8080, self.cai_enabled)
                logger.debug(
                    "[refresh-server] server_num:%s server_list:%s" % (len(server_list), server_list))
                if not server_list:
                    logger.error("[refresh-server] empty server_list get from %s, do not refresh" % self.endpoint)
                    continue
                with self.server_list_lock:
                    self.server_list = server_list
                    self.server_offset = 0
                    if self.current_server not in server_list:
                        logger.warning("[refresh-server] %s is not effective, change one" % str(self.current_server))
                        self.current_server = server_list[self.server_offset]
            except Exception as e:
                logger.exception("[refresh-server] exception %s occur" % str(e))

    def change_server(self):
        with self.server_list_lock:
            self.server_offset = (self.server_offset + 1) % len(self.server_list)
            self.current_server = self.server_list[self.server_offset]

    def get_server(self):
        if self.server_list is None:
            with self.server_list_lock:
                logger.info("[get-server] server list is null, try to initialize")
                server_list = get_server_list(self.endpoint, 443 if self.tls_enabled else 8080, self.cai_enabled)
                if not server_list:
                    logger.error("[get-server] empty server_list get from %s" % self.endpoint)
                    return None
                self.server_list = server_list
                self.current_server = self.server_list[self.server_offset]
                logger.info("[get-server] server_num:%s server_list:%s" % (len(self.server_list), self.server_list))

            if self.cai_enabled:
                t = Thread(target=self._refresh_server_list)
                t.setDaemon(True)
                t.start()

        logger.info("[get-server] use server:%s" % str(self.current_server))
        return self.current_server

    def remove(self, data_id, group, timeout=None):
        """ Remove one data item from ACM.

        :param data_id: dataId.
        :param group: group, use "DEFAULT_GROUP" if no group specified.
        :param timeout: timeout for requesting server in seconds.
        :return: True if success or an exception will be raised.
        """
        data_id, group = process_common_params(data_id, group)
        logger.info(
            "[remove] data_id:%s, group:%s, namespace:%s, timeout:%s" % (data_id, group, self.namespace, timeout))

        params = {
            "dataId": data_id,
            "group": group,
        }
        if self.namespace:
            params["tenant"] = self.namespace

        try:
            resp = self._do_sync_req("/diamond-server/datum.do?method=deleteAllDatums", None, None, params,
                                     timeout or self.default_timeout)
            logger.info("[remove] success to remove group:%s, data_id:%s, server response:%s" % (
                group, data_id, resp.read()))
            return True
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                logger.error(
                    "[remove] no right for namespace:%s, group:%s, data_id:%s" % (self.namespace, group, data_id))
                raise ACMException("Insufficient privilege.")
            else:
                logger.error("[remove] error code [:%s] for namespace:%s, group:%s, data_id:%s" % (
                    e.code, self.namespace, group, data_id))
                raise ACMException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[remove] exception %s occur" % str(e))
            raise

    def publish(self, data_id, group, content, timeout=None):
        """ Publish one data item to ACM.

        If the data key is not exist, create one first.
        If the data key is exist, update to the content specified.
        Content can not be set to None, if there is need to delete config item, use function **remove** instead.

        :param data_id: dataId.
        :param group: group, use "DEFAULT_GROUP" if no group specified.
        :param content: content of the data item.
        :param timeout: timeout for requesting server in seconds.
        :return: True if success or an exception will be raised.
        """
        if content is None:
            raise ACMException("Can not publish none content, use remove instead.")

        data_id, group = process_common_params(data_id, group)
        if type(content) == bytes:
            content = content.decode("utf8")

        if is_encrypted(data_id) and self.kms_enabled:
            content = self.encrypt(content)

        logger.info("[publish] data_id:%s, group:%s, namespace:%s, content:%s, timeout:%s" % (
            data_id, group, self.namespace, truncate(content), timeout))
        params = {
            "dataId": data_id,
            "group": group,
            "content": content.encode("GBK"),
        }
        if self.namespace:
            params["tenant"] = self.namespace

        try:
            resp = self._do_sync_req("/diamond-server/basestone.do?method=syncUpdateAll", None, None, params,
                                     timeout or self.default_timeout)
            logger.info("[publish] success to publish content, group:%s, data_id:%s, server response:%s" % (
                group, data_id, resp.read()))
            return True
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                logger.error(
                    "[publish] no right for namespace:%s, group:%s, data_id:%s" % (self.namespace, group, data_id))
                raise ACMException("Insufficient privilege.")
            else:
                logger.error("[publish] error code [:%s] for namespace:%s, group:%s, data_id:%s" % (
                    e.code, self.namespace, group, data_id))
                raise ACMException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[publish] exception %s occur" % str(e))
            raise

    def get(self, data_id, group, timeout=None, no_snapshot=False):
        content = self.get_raw(data_id, group, timeout, no_snapshot)
        if content and is_encrypted(data_id) and self.kms_enabled:
            return self.decrypt(content)
        return content

    def get_raw(self, data_id, group, timeout=None, no_snapshot=False):
        """Get value of one config item.

        Query priority:
        1.  Get from local failover dir(default: "{cwd}/acm/data").
            Failover dir can be manually copied from snapshot dir(default: "{cwd}/acm/snapshot") in advance.
            This helps to suppress the effect of known server failure.

        2.  Get from one server until value is got or all servers tried.
            Content will be save to snapshot dir.

        3.  Get from snapshot dir.

        :param data_id: dataId.
        :param group: group, use "DEFAULT_GROUP" if no group specified.
        :param timeout: timeout for requesting server in seconds.
        :param no_snapshot: do not save snapshot.
        :return: value.
        """
        data_id, group = process_common_params(data_id, group)
        logger.info("[get-config] data_id:%s, group:%s, namespace:%s, timeout:%s" % (
            data_id, group, self.namespace, timeout))

        params = {
            "dataId": data_id,
            "group": group,
        }
        if self.namespace:
            params["tenant"] = self.namespace

        cache_key = group_key(data_id, group, self.namespace)
        # get from failover
        content = read_file(self.failover_base, cache_key)
        if content is None:
            logger.debug("[get-config] failover config is not exist for %s, try to get from server" % cache_key)
        else:
            logger.debug("[get-config] get %s from failover directory, content is %s" % (cache_key, truncate(content)))
            return content

        # get from server
        try:
            resp = self._do_sync_req("/diamond-server/config.co", None, params, None, timeout or self.default_timeout)
            content = resp.read().decode("GBK")
        except HTTPError as e:
            if e.code == HTTPStatus.NOT_FOUND:
                logger.warning(
                    "[get-config] config not found for data_id:%s, group:%s, namespace:%s, try to delete snapshot" % (
                        data_id, group, self.namespace))
                delete_file(self.snapshot_base, cache_key)
                return None
            elif e.code == HTTPStatus.CONFLICT:
                logger.error(
                    "[get-config] config being modified concurrently for data_id:%s, group:%s, namespace:%s" % (
                        data_id, group, self.namespace))
            elif e.code == HTTPStatus.FORBIDDEN:
                logger.error("[get-config] no right for data_id:%s, group:%s, namespace:%s" % (
                    data_id, group, self.namespace))
                raise ACMException("Insufficient privilege.")
            else:
                logger.error("[get-config] error code [:%s] for data_id:%s, group:%s, namespace:%s" % (
                    e.code, data_id, group, self.namespace))
                if no_snapshot:
                    raise
        except Exception as e:
            logger.exception("[get-config] exception %s occur" % str(e))
            if no_snapshot:
                raise

        if no_snapshot:
            return content

        if content is not None:
            logger.info(
                "[get-config] content from server:%s, data_id:%s, group:%s, namespace:%s, try to save snapshot" % (
                    truncate(content), data_id, group, self.namespace))
            try:
                save_file(self.snapshot_base, cache_key, content)
            except Exception as e:
                logger.exception("[get-config] save snapshot failed for %s, data_id:%s, group:%s, namespace:%s" % (
                    data_id, group, self.namespace, str(e)))
            return content

        logger.error("[get-config] get config from server failed, try snapshot, data_id:%s, group:%s, namespace:%s" % (
            data_id, group, self.namespace))
        content = read_file(self.snapshot_base, cache_key)
        if content is None:
            logger.warning("[get-config] snapshot is not exist for %s." % cache_key)
        else:
            logger.debug("[get-config] get %s from snapshot directory, content is %s" % (cache_key, truncate(content)))
            return content

    def list(self, page=1, size=200):
        """ Get config items of current namespace with content included.

        Data is directly from acm server.

        :param page: which page to query, starts from 1.
        :param size: page size.
        :return:
        """
        logger.info("[list] try to list namespace:%s" % self.namespace)

        params = {
            "pageNo": page,
            "pageSize": size,
            "method": "getAllConfigByTenant",
        }

        if self.namespace:
            params["tenant"] = self.namespace

        try:
            resp = self._do_sync_req("/diamond-server/basestone.do", None, params, None, self.default_timeout)
            d = resp.read()
            return json.loads(d)
        except HTTPError as e:
            if e.code == HTTPStatus.FORBIDDEN:
                logger.error("[list] no right for namespace:%s" % self.namespace)
                raise ACMException("Insufficient privilege.")
            else:
                logger.error("[list] error code [%s] for namespace:%s" % (e.code, self.namespace))
                raise ACMException("Request Error, code is %s" % e.code)
        except Exception as e:
            logger.exception("[list] exception %s occur" % str(e))
            raise

    def list_all(self, group=None, prefix=None):
        """ Get all config items of current namespace, with content included.

        Warning: If there are lots of config in namespace, this function may cost some time.

        :param group: only dataIds with group match shall be returned.
        :param prefix: only dataIds startswith prefix shall be returned **it's case sensitive**.
        :return:
        """
        logger.info("[list-all] namespace:%s, group:%s, prefix:%s" % (self.namespace, group, prefix))

        def _matches(ori):
            return (group is None or ori["group"] == group) and (prefix is None or ori["dataId"].startswith(prefix))

        result = self.list(1, 200)
        if not result:
            logger.warning("[list-all] can not get config items of %s" % self.namespace)
            return list()

        ret_list = [{"dataId": i["dataId"], "group": i["group"]} for i in result["pageItems"] if _matches(i)]
        pages = result["pagesAvailable"]
        logger.debug("[list-all] %s items got from acm server" % result["totalCount"])

        for i in range(2, pages + 1):
            result = self.list(i, 200)
            ret_list += [{"dataId": j["dataId"], "group": j["group"]} for j in result["pageItems"] if _matches(j)]
        logger.debug("[list-all] %s items returned" % len(ret_list))
        return ret_list

    @synchronized_with_attr("pulling_lock")
    def add_watcher(self, data_id, group, cb):
        self.add_watchers(data_id, group, [cb])

    @synchronized_with_attr("pulling_lock")
    def add_watchers(self, data_id, group, cb_list):
        """Add watchers to specified item.

        1.  Callback is invoked from current process concurrently by thread pool.
        2.  Callback is invoked at once if the item exists.
        3.  Callback is invoked if changes or deletion detected on the item.

        :param data_id: dataId.
        :param group: group, use "DEFAULT_GROUP" if no group specified.
        :param cb_list: callback functions.
        :return:
        """
        if not cb_list:
            raise ACMException("A callback function is needed.")
        data_id, group = process_common_params(data_id, group)
        logger.info("[add-watcher] data_id:%s, group:%s, namespace:%s" % (data_id, group, self.namespace))
        cache_key = group_key(data_id, group, self.namespace)
        wl = self.watcher_mapping.get(cache_key)
        if not wl:
            wl = list()
            self.watcher_mapping[cache_key] = wl
        for cb in cb_list:
            wl.append(WatcherWrap(cache_key, cb))
            logger.info("[add-watcher] watcher has been added for key:%s, new callback is:%s, callback number is:%s" % (
                cache_key, cb.__name__, len(wl)))

        if self.puller_mapping is None:
            logger.debug("[add-watcher] pulling should be initialized")
            self._int_pulling()

        if cache_key in self.puller_mapping:
            logger.debug("[add-watcher] key:%s is already in pulling" % cache_key)
            return

        for key, puller_info in self.puller_mapping.items():
            if len(puller_info[1]) < self.pulling_config_size:
                logger.debug("[add-watcher] puller:%s is available, add key:%s" % (puller_info[0], cache_key))
                puller_info[1].append(key)
                self.puller_mapping[cache_key] = puller_info
                break
        else:
            logger.debug("[add-watcher] no puller available, new one and add key:%s" % cache_key)
            key_list = self.process_mgr.list()
            key_list.append(cache_key)
            puller = Process(target=self._do_pulling, args=(key_list, self.notify_queue))
            puller.daemon = True
            puller.start()
            self.puller_mapping[cache_key] = (puller, key_list)

    @synchronized_with_attr("pulling_lock")
    def remove_watcher(self, data_id, group, cb, remove_all=False):
        """Remove watcher from specified key.

        :param data_id: dataId.
        :param group: group, use "DEFAULT_GROUP" if no group specified.
        :param cb: callback function.
        :param remove_all: weather to remove all occurrence of the callback or just once.
        :return:
        """
        if not cb:
            raise ACMException("A callback function is needed.")
        data_id, group = process_common_params(data_id, group)
        if not self.puller_mapping:
            logger.warning("[remove-watcher] watcher is never started.")
            return
        cache_key = group_key(data_id, group, self.namespace)
        wl = self.watcher_mapping.get(cache_key)
        if not wl:
            logger.warning("[remove-watcher] there is no watcher on key:%s" % cache_key)
            return

        wrap_to_remove = list()
        for i in wl:
            if i.callback == cb:
                wrap_to_remove.append(i)
                if not remove_all:
                    break

        for i in wrap_to_remove:
            wl.remove(i)

        logger.info("[remove-watcher] %s is removed from %s, remove all:%s" % (cb.__name__, cache_key, remove_all))
        if not wl:
            logger.debug("[remove-watcher] there is no watcher for:%s, kick out from pulling" % cache_key)
            self.watcher_mapping.pop(cache_key)
            puller_info = self.puller_mapping[cache_key]
            puller_info[1].remove(cache_key)
            if not puller_info[1]:
                logger.debug("[remove-watcher] there is no pulling keys for puller:%s, stop it" % puller_info[0])
                self.puller_mapping.pop(cache_key)
                puller_info[0].terminate()

    def _do_sync_req(self, url, headers=None, params=None, data=None, timeout=None):
        url = "?".join([url, urlencode(params)]) if params else url
        all_headers = self._get_common_headers(params, data)
        if headers:
            all_headers.update(headers)
        logger.debug(
            "[do-sync-req] url:%s, headers:%s, params:%s, data:%s, timeout:%s" % (
                url, all_headers, params, data, timeout))
        tries = 0
        while True:
            try:
                server_info = self.get_server()
                if not server_info:
                    logger.error("[do-sync-req] can not get one server.")
                    raise ACMRequestException("Server is not available.")
                address, port, is_ip_address = server_info
                server = ":".join([address, str(port)])
                # if tls is enabled and server address is in ip, turn off verification

                server_url = "%s://%s" % ("https" if self.tls_enabled else "http", server)
                req = Request(url=server_url + url, data=urlencode(data).encode() if data else None,
                              headers=all_headers)

                # for python2.6 compatibility
                if sys.version_info[0] == 2 and sys.version_info[1] == 6:
                    resp = urlopen(req, timeout=timeout)
                else:
                    if self.tls_enabled and is_ip_address:
                        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                        context.check_hostname = False
                    else:
                        context = None
                    resp = urlopen(req, timeout=timeout, context=context)
                logger.debug("[do-sync-req] info from server:%s" % server)
                return resp
            except HTTPError as e:
                if e.code in [HTTPStatus.INTERNAL_SERVER_ERROR, HTTPStatus.BAD_GATEWAY,
                              HTTPStatus.SERVICE_UNAVAILABLE]:
                    logger.warning("[do-sync-req] server:%s is not available for reason:%s" % (server, e.msg))
                else:
                    raise
            except socket.timeout:
                logger.warning("[do-sync-req] %s request timeout" % server)
            except URLError as e:
                logger.warning("[do-sync-req] %s connection error:%s" % (server, e.reason))

            tries += 1
            if tries >= len(self.server_list):
                logger.error("[do-sync-req] %s maybe down, no server is currently available" % server)
                raise ACMRequestException("All server are not available")
            self.change_server()
            logger.warning("[do-sync-req] %s maybe down, skip to next" % server)

    def _do_pulling(self, cache_list, queue):
        cache_pool = dict()
        for cache_key in cache_list:
            cache_pool[cache_key] = CacheData(cache_key, self)

        while cache_list:
            unused_keys = set(cache_pool.keys())
            contains_init_key = False
            probe_update_string = ""
            for cache_key in cache_list:
                cache_data = cache_pool.get(cache_key)
                if not cache_data:
                    logger.debug("[do-pulling] new key added: %s" % cache_key)
                    cache_data = CacheData(cache_key, self)
                    cache_pool[cache_key] = cache_data
                if cache_data.is_init:
                    contains_init_key = True
                data_id, group, namespace = parse_key(cache_key)
                probe_update_string += WORD_SEPARATOR.join(
                    [data_id, group, cache_data.md5 or "", self.namespace]) + LINE_SEPARATOR
                unused_keys.remove(cache_key)
            for k in unused_keys:
                logger.debug("[do-pulling] %s is no longer watched, remove from cache" % k)
                cache_pool.pop(k)

            logger.debug(
                "[do-pulling] try to detected change from server probe string is %s" % truncate(probe_update_string))
            headers = {"longPullingTimeout": int(self.pulling_timeout * 1000)}
            if contains_init_key:
                headers["longPullingNoHangUp"] = "true"

            data = {"Probe-Modify-Request": probe_update_string}

            changed_keys = list()
            try:
                resp = self._do_sync_req("/diamond-server/config.co", headers, None, data, self.pulling_timeout + 10)
                changed_keys = [group_key(*i) for i in parse_pulling_result(resp.read())]
                logger.debug("[do-pulling] following keys are changed from server %s" % truncate(str(changed_keys)))
            except ACMException as e:
                logger.error("[do-pulling] acm exception: %s" % str(e))
            except Exception as e:
                logger.exception("[do-pulling] exception %s occur, return empty list" % str(e))

            for cache_key, cache_data in cache_pool.items():
                cache_data.is_init = False
                if cache_key in changed_keys:
                    data_id, group, namespace = parse_key(cache_key)
                    content = self.get(data_id, group)
                    md5 = hashlib.md5(content.encode("GBK")).hexdigest() if content is not None else None
                    cache_data.md5 = md5
                    cache_data.content = content
                queue.put((cache_key, cache_data.content, cache_data.md5))

    @synchronized_with_attr("pulling_lock")
    def _int_pulling(self):
        if self.puller_mapping is not None:
            logger.info("[init-pulling] puller is already initialized")
            return
        self.puller_mapping = dict()
        self.notify_queue = Queue()
        self.callback_tread_pool = pool.ThreadPool(self.callback_tread_num)
        self.process_mgr = Manager()
        t = Thread(target=self._process_polling_result)
        t.setDaemon(True)
        t.start()
        logger.info("[init-pulling] init completed")

    def _process_polling_result(self):
        while True:
            cache_key, content, md5 = self.notify_queue.get()
            logger.debug("[process-polling-result] receive an event:%s" % cache_key)
            wl = self.watcher_mapping.get(cache_key)
            if not wl:
                logger.warning("[process-polling-result] no watcher on %s, ignored" % cache_key)
                continue

            data_id, group, namespace = parse_key(cache_key)
            params = {
                "data_id": data_id,
                "group": group,
                "namespace": namespace,
                "content": content
            }
            for watcher in wl:
                if not watcher.last_md5 == md5:
                    logger.debug(
                        "[process-polling-result] md5 changed since last call, calling %s" % watcher.callback.__name__)
                    try:
                        self.callback_tread_pool.apply(watcher.callback, (params,))
                    except Exception as e:
                        logger.exception("[process-polling-result] exception %s occur while calling %s " % (
                            str(e), watcher.callback.__name__))
                    watcher.last_md5 = md5

    def _get_common_headers(self, params, data):
        headers = {
            "Diamond-Client-AppName": self.app_name,
            "Client-Version": VERSION,
            "exConfigInfo": "true",
        }
        if data:
            headers["Content-Type"] = "application/x-www-form-urlencoded; charset=GBK"

        if self.auth_enabled:
            ts = str(int(time.time() * 1000))
            headers.update({
                "Spas-AccessKey": self.ak,
                "timeStamp": ts,
            })
            sign_str = ""
            # in case tenant or group is null
            if not params and not data:
                return headers

            tenant = (params and params.get("tenant")) or (data and data.get("tenant"))
            group = (params and params.get("group")) or (data and data.get("group"))

            if tenant:
                sign_str = tenant + "+"
            if group:
                sign_str = sign_str + group + "+"

            if sign_str:
                sign_str += ts
                headers["Spas-Signature"] = base64.encodebytes(
                    hmac.new(self.sk.encode(), sign_str.encode(), digestmod=hashlib.sha1).digest()).decode().strip()
        return headers

    def _prepare_kms(self):
        if not (self.region_id and self.kms_ak and self.kms_secret):
            return False
        if not self.kms_client:
            self.kms_client = AcsClient(ak=self.kms_ak, secret=self.kms_secret, region_id=self.region_id)
        return True

    def encrypt(self, plain_txt):
        if not self._prepare_kms():
            return plain_txt
        ssl._create_default_https_context = ssl._create_unverified_context
        req = EncryptRequest()
        req.set_KeyId(self.key_id)
        req.set_Plaintext(plain_txt if type(plain_txt) == bytes else plain_txt.encode("utf8"))
        resp = json.loads(self.kms_client.do_action_with_exception(req))
        return resp["CiphertextBlob"]

    def decrypt(self, cipher_blob):
        if not self._prepare_kms():
            return cipher_blob
        ssl._create_default_https_context = ssl._create_unverified_context
        req = DecryptRequest()
        req.set_CiphertextBlob(cipher_blob)
        resp = json.loads(self.kms_client.do_action_with_exception(req))
        return resp["Plaintext"]


if DEBUG:
    ACMClient.set_debugging()



import socket
import logging
import random

try:
    # python3.6
    from urllib.request import Request, urlopen
    from urllib.error import URLError
except ImportError:
    # python2.7
    from urllib2 import Request, urlopen, URLError

logger = logging.getLogger("acm")

ADDRESS_URL_PTN = "http://%s/diamond-server/diamond"

ADDRESS_SERVER_TIMEOUT = 3  # in seconds


def is_ipv4_address(address):
    try:
        socket.inet_aton(address)
    except socket.error:
        return False
    return True


def get_server_list(endpoint, default_port=8080, cai_enabled=True):
    server_list = list()
    if not cai_enabled:
        logger.info("[get-server-list] cai server is not used, regard endpoint:%s as server." % endpoint)
        content = endpoint.encode()
    else:
        try:
            # use 8080 as default port.
            if ":" not in endpoint:
                endpoint = endpoint + ":8080"
            content = urlopen(ADDRESS_URL_PTN % endpoint, timeout=ADDRESS_SERVER_TIMEOUT).read()
            logger.debug("[get-server-list] content from endpoint:%s" % content)
        except (URLError, OSError, socket.timeout) as e:
            logger.error("[get-server-list] get server from %s failed, cause:%s" % (endpoint, e))
            return server_list

    if content:
        for server_info in content.decode().strip().split("\n"):
            sp = server_info.strip().split(":")
            if len(sp) == 1:
                server_list.append((sp[0], default_port, is_ipv4_address(sp[0])))
            else:
                try:
                    server_list.append((sp[0], int(sp[1]), is_ipv4_address(sp[0])))
                except ValueError:
                    logger.warning("[get-server-list] bad server address:%s ignored" % server_info)

    random.shuffle(server_list)

    return server_list



import os.path
import sys
import json
import fcntl
import shutil
import gettext
from datetime import datetime
import zipfile
from acm import ACMClient, DEFAULT_GROUP_NAME


# override the default expression of "positional arguments"
def translate_patch(msg):
    return "required arguments" if msg == "positional arguments" else msg


gettext.gettext = translate_patch

import argparse

###

DEFAULT_ENDPOINT = "acm.aliyun.com"
DEFAULT_PORT = 8080
CMD = set(["bind", "set", "use"])
CONF = os.path.join(os.getenv("HOME"), ".acm.json")

INIT_CONF = {
    "endpoints": {
        DEFAULT_ENDPOINT: {
            "tls": False,
            "is_current": True,
            "region_id": None,
            "kms_enabled": False,
            "namespaces": {
                "[default]": {
                    "is_current": True,
                    "ak": None,
                    "sk": None,
                    "alias": "[default]",
                    "kms_ak": None,
                    "kms_secret": None,
                    "key_id": None,
                    "updated": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                }
            }
        }
    }
}


def _colored(txt, color="green"):
    cm = {
        "green": 32,
        "red": 31,
        "yellow": 33,
        "grey": 30,
    }

    return "\033[1;%sm%s\033[0m" % (cm[color], txt)


def read_config():
    try:
        if not os.path.exists(CONF):
            write_config(INIT_CONF)
        if sys.version_info[0] == 3:
            with open(CONF, "r+", newline="") as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                return json.loads(f.read())
        else:
            with open(CONF, "r+") as f:
                fcntl.flock(f, fcntl.LOCK_EX)
                return json.loads(f.read())
    except Exception as e:
        print("Read config error due to %s" % str(e))
        sys.exit(1)


def write_config(config):
    try:
        content = json.dumps(config, indent=4)
        with open(CONF, "wb") as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            f.write(content if type(content) == bytes else content.encode("utf8"))
    except Exception as e:
        print("Save config error due to %s" % str(e))
        sys.exit(1)


def _get_current(config):
    endpoint = ""
    namespace = ""
    for k, v in config["endpoints"].items():
        if not v["is_current"]:
            continue
        endpoint = k
        for k2, v2 in v["namespaces"].items():
            if not v2["is_current"]:
                continue
            namespace = k2
            break

    return endpoint, namespace


def _set_current(config, endpoint, namespace=None):
    for k, v in config["endpoints"].items():
        if k == endpoint:
            v["is_current"] = True
            if namespace is not None:
                for k2, v2 in v["namespaces"].items():
                    if k2 == namespace:
                        v2["is_current"] = True
                    else:
                        v2["is_current"] = False
        else:
            v["is_current"] = False

    return config


def add(args):
    if ":" in args.namespace:
        pos = args.namespace.index(":")
        e = args.namespace[:pos]
        ns = args.namespace[pos + 1:]
    else:
        e = DEFAULT_ENDPOINT
        ns = args.namespace
    config = read_config()
    alias = args.alias or ns

    if args.alias is not None and ":" in args.alias:
        print('":" is invalid symbol in alias.')
        sys.exit(1)

    # detect alias, ensure unique globally
    for ep, ep_info in config["endpoints"].items():
        for k, v in ep_info["namespaces"].items():
            if args.alias is None and v["alias"] == alias and (k != ns or ep != e):
                alias = "-".join([e, ns])
            elif v["alias"] == alias and k != ns:
                print("Alias %s has been taken by %s:%s, choose another one." % (_colored(alias, "red"), ep, k))
                sys.exit(1)

    # new endpoint
    if e not in config["endpoints"]:
        if args.kms:
            if not args.region_id:
                print(_colored("Region ID", "red") + " must be specified to use KMS.")
                sys.exit(1)
        config["endpoints"][e] = {
            "tls": args.tls,
            "is_current": False,
            "region_id": args.region_id,
            "kms_enabled": args.kms,
            "namespaces": {}
        }
        print(
                "Adding a new endpoint: %s, using TLS is %s.\n" % (_colored(e, "yellow"), _colored(args.tls, "yellow")))
    else:
        endpoint = config["endpoints"][e]
        if args.kms and not args.region_id and not endpoint.get("region_id"):
            print(_colored("Region ID", "red") + " must be specified to use KMS.")
            sys.exit(1)
        if endpoint.get("tls") != args.tls:
            endpoint["tls"] = args.tls
            print("TLS attr of %s has changed to %s.\n" % (_colored(e, "yellow"), _colored(args.tls, "yellow")))
        if endpoint.get("kms_enabled") != args.kms:
            endpoint["kms_enabled"] = args.kms
            print("KMS enabled of %s has turned to %s.\n" % (_colored(e, "yellow"), _colored(args.kms, "yellow")))
        if args.region_id is not None:
            if endpoint.get("region_id") != args.region_id:
                endpoint["region_id"] = args.region_id
                print("Region ID of %s has changed to %s.\n" % (
                _colored(e, "yellow"), _colored(args.region_id, "yellow")))

    if ns in config["endpoints"][e]["namespaces"]:
        namespace = config["endpoints"][e]["namespaces"][ns]
        if args.ak is not None:
            namespace["ak"] = args.ak
        if args.sk is not None:
            namespace["sk"] = args.sk
        if args.alias is not None:
            namespace["alias"] = alias
        if args.kms_ak is not None:
            namespace["kms_ak"] = args.kms_ak
        if args.kms_secret is not None:
            namespace["kms_secret"] = args.kms_secret
        if args.key_id is not None:
            namespace["key_id"] = args.key_id
        if args.kms:
            if not namespace.get("kms_ak"):
                if not namespace.get("ak"):
                    print(_colored("AccessKey", "red") + ' must be specified to use KMS.')
                    sys.exit(1)
                namespace["kms_ak"] = namespace.get("ak")
            if not namespace.get("kms_secret"):
                if not namespace.get("sk"):
                    print(_colored("SecretKey", "red") + ' must be specified to use KMS.')
                    sys.exit(1)
                namespace["kms_secret"] = namespace.get("sk")
        namespace["updated"] = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print("Namespace %s is already exist in %s, updating configs.\n" % (
            _colored(ns, "green"), _colored(e, "yellow")))
    else:
        namespace = {"ak": args.ak, "sk": args.sk, "alias": alias, "is_current": False,
                     "updated": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                     "kms_ak": None, "kms_secret": None, "key_id": None}
        if args.kms:
            kms_ak = args.kms_ak or args.ak
            if not kms_ak:
                print(_colored("AccessKey", "red") + ' must be specified to use KMS.')
                sys.exit(1)
            kms_secret = args.kms_secret or args.sk
            if not kms_secret:
                print(_colored("SecretKey", "red") + ' must be specified to use KMS.')
                sys.exit(1)
            namespace["kms_ak"] = kms_ak
            namespace["kms_secret"] = kms_secret
            namespace["key_id"] = args.key_id
        config["endpoints"][e]["namespaces"][ns] = namespace
        print(
                "Add new namespace %s(%s) to %s.\n" %
                (_colored(ns, "green"), _colored(alias, "green"), _colored(e, "yellow")))

    write_config(config)

    try:
        print("Try to access the namespace...")
        c = ACMClient(endpoint=e, namespace=(None if ns == "[default]" else ns),
                      ak=config["endpoints"][e]["namespaces"][ns]["ak"],
                      sk=config["endpoints"][e]["namespaces"][ns]["sk"])
        if config["endpoints"][e]["tls"]:
            c.set_options(tls_enabled=True)
        c.list(1, 1)
        print("Namespace access succeed.")
    except:
        print(_colored("\nWarning: Access test failed, there may be mistakes in configuration.\n", "grey"))


def use(args):
    config = read_config()

    if ":" in args.namespace:
        pos = args.namespace.index(":")
        e = args.namespace[:pos]
        ns = args.namespace[pos + 1:]
    else:
        e = None
        ns = None

    found = False
    # detect alias, ensure unique globally
    for ep, ep_info in config["endpoints"].items():
        for k, v in ep_info["namespaces"].items():
            if v["alias"] == args.namespace or (k == ns and ep == e):
                _set_current(config, ep, k)
                print("Namespace changed to %s alias:%s.\n" % (
                    _colored("%s:%s" % (ep, k), "green"), _colored(v["alias"], "green")))
                write_config(config)
                found = True
                break
        if found:
            break
    else:
        print("No namespace named or aliased as %s, please check.\n" % _colored(args.namespace, "red"))


def _process_namespace(args):
    config = read_config()

    if args.namespace is not None:
        if ":" in args.namespace:
            pos = args.namespace.index(":")
            e = args.namespace[:pos]
            ns = args.namespace[pos + 1:]
        else:
            e = None
            ns = None
        for ep, ep_info in config["endpoints"].items():
            for k, v in ep_info["namespaces"].items():
                if v["alias"] == args.namespace or (k == ns and ep == e):
                    return ep, ep_info, k, v
        print("No namespace named or aliased as %s, please check.\n" % _colored(args.namespace, "red"))
        sys.exit(1)

    e, n = _get_current(config)
    return e, config["endpoints"][e], n, config["endpoints"][e]["namespaces"][n]


def list_conf(args):
    e, ep, n, ns = _process_namespace(args)
    c = ACMClient(endpoint=e, namespace=(None if n == "[default]" else n), ak=ns["ak"], sk=ns["sk"])
    if ep["tls"]:
        c.set_options(tls_enabled=True)

    try:
        configs = c.list_all(args.group, args.prefix)
    except:
        print("List failed.")
        sys.exit(1)

    for i in sorted(configs, key=lambda x: x["group"] + x["dataId"]):
        print("%(group)s/%(dataId)s" % i)


def _write_file(file, content):
    try:
        with open(file, "wb") as f:
            f.write(content if type(content) == bytes else content.encode("utf8"))
    except Exception as e:
        print("Write file error due to %s" % str(e))
        sys.exit(1)


def _read_file(file):
    try:
        if sys.version_info[0] == 3:
            with open(file, "r+", newline="") as f:
                return f.read()
        else:
            with open(file, "r+") as f:
                return f.read()
    except Exception as e:
        print("Read file error due to %s" % str(e))
        sys.exit(1)


def pull(args):
    e, ep, n, ns = _process_namespace(args)
    c = _get_client(e, ep, n, ns)
    try:
        if "/" in args.data_id:
            g, d = args.data_id.split("/")
            content = c.get(d, g, no_snapshot=True)
        else:
            content = c.get(args.data_id, None, no_snapshot=True)
    except:
        print("Pull %s failed." % args.data_id)
        sys.exit(1)

    if content is None:
        print("%s does not exist." % _colored(args.data_id, "red"))
        sys.exit(1)
    os.write(1, content.encode("utf8"))


def push(args):
    if args.file:
        if not sys.stdin.isatty():
            print(_colored("Warning: content from stdin will be ignored since file is specified.", "grey"))
        if not os.path.exists(args.file):
            print("File %s does not exist." % _colored(args.file, "red"))
            sys.exit(1)
        content = _read_file(args.file)
    elif not sys.stdin.isatty():
        content = sys.stdin.read()
    else:
        print("Use file or stdin as input.")
        sys.exit(1)

    e, ep, n, ns = _process_namespace(args)
    c = _get_client(e, ep, n, ns)

    if args.data_id.count("/") > 1:
        print("Invalid dataId or filename, more than one / is given.")
        sys.exit(1)

    group, data_id = args.data_id.split("/") if "/" in args.data_id else (None, args.data_id)

    try:
        c.publish(data_id, group, content)
    except:
        import traceback
        traceback.print_exc()
        print("Push %s failed." % args.data_id)
        sys.exit(1)

    print("content has been pushed to [dataId:%s].\n" % (_colored(args.data_id, "green")))


def current(args):
    config = read_config()
    e, n = _get_current(config)
    print("Current Endpoint:\t%s, using TLS is %s, using KMS is %s, Region ID is %s." % (
        _colored(e, "yellow"), _colored(config["endpoints"][e].get("tls"), "yellow"),
        _colored(config["endpoints"][e].get("kms_enabled"), "yellow"),
        _colored(config["endpoints"][e].get("region_id"), "yellow")))
    print("\nCurrent Namespace:\t%s" % (_colored(n)))
    ns = config["endpoints"][e]["namespaces"][n]

    print("\tAlias:\t\t%s" % _colored(ns["alias"], "green"))
    print("\tAccessKey:\t%s" % ns["ak"])
    print("\tSecretKey:\t%s" % ns["sk"])
    if config["endpoints"][e].get("kms_enabled"):
        print("\tAccessKey for KMS:\t%s" % ns.get("kms_ak"))
        print("\tSecretKey for KMS:\t%s" % ns.get("kms_secret"))
        print("\tKey ID:\t%s" % ns.get("key_id"))
    print("")


def show(args):
    config = read_config()
    e, n = _get_current(config)
    max_ep = 10
    max_ns = 10
    max_alias = 10

    table_header = ["", "ENDPOINT", "NAMESPACE_ID", "ALIAS", "UPDATED"]
    table_data = list()

    for k, v in config["endpoints"].items():
        if len(k) > max_ep:
            max_ep = len(k)
        start = True
        for k2, v2 in v["namespaces"].items():
            if k == e and k2 == n:
                row_data = ["*"]
            else:
                row_data = [""]
            if start:
                row_data.append(k)
                start = False
            else:
                row_data.append(k)
            if len(k2) > max_ns:
                max_ns = len(k2)
            if len(v2["alias"]) > max_alias:
                max_alias = len(v2["alias"])
            row_data.append(k2)
            row_data.append(v2["alias"])
            row_data.append(v2.get("updated", "None"))
            table_data.append(row_data)
    table_data = sorted(table_data, key=lambda x: x[4], reverse=True)
    ptn = "%%-3s%%-%is%%-%is%%-%is%%-20s" % (max_ep + 5, max_ns + 5, max_alias + 5)
    print(ptn % tuple(table_header))
    print("-" * (max_ep + max_ns + max_alias + 38))
    for row in table_data:
        print(ptn % tuple(row))
    print("-" * (max_ep + max_ns + max_alias + 38))
    print("")


def export(args):
    e, ep, n, ns = _process_namespace(args)
    c = _get_client(e, ep, n, ns)
    try:
        configs = c.list_all()
    except:
        print("Get config list failed.")
        sys.exit(1)

    groups = set()
    elements = set()
    for i in configs:
        groups.add(i["group"])
        elements.add(os.path.join(i["group"], i["dataId"]) if i["group"] != DEFAULT_GROUP_NAME else i["dataId"])
    dest_file = args.file or "%s-%s.zip" % (e, n)
    zip_file = None

    if args.dir:
        try:
            os.makedirs(args.dir)
        except OSError:
            pass
        # process deleting
        if args.delete:
            candidates = list()
            # get candidates
            for root, dirs, files in os.walk(args.dir):
                if not os.path.basename(root).startswith("."):
                    for i in dirs:
                        if i.startswith("."):
                            continue

                        if i not in groups:
                            candidates.append(os.path.join(root, i))

                    for i in files:
                        if i.startswith("."):
                            continue
                        candidates.append(os.path.join(root, i))
            # kick out elements
            delete_list = list()
            trunc_len = len(args.dir) + len(os.path.sep)
            for i in candidates:
                if i[trunc_len:] not in elements:
                    delete_list.append(i)

            # deleting
            if delete_list:
                print("Following files and dirs are not exist in ACM Server:\n")
                for i in delete_list:
                    print(" - " + i)

                delete = True
                if not args.force:
                    while True:
                        if sys.version_info[0] == 3:
                            choice = input("\nDeleting all files above? (y/n)")
                        else:
                            choice = raw_input("\nDeleting all files above? (y/n)")
                        if choice.lower() in ["y", "n"]:
                            delete = choice.lower() == "y"
                            break
                        print("Invalid choice, please input y or n.")
                if delete:
                    for i in delete_list:
                        try:
                            if os.path.isfile(i):
                                os.remove(i)
                            else:
                                shutil.rmtree(i)
                        except OSError:
                            pass
                    print("Delete complete, continue to export...\n")
    else:
        zip_file = zipfile.ZipFile(dest_file, 'w', zipfile.ZIP_DEFLATED)

    print(_colored(len(configs), "green") + " dataIds on ACM server will be exported to %s.\n" % _colored(
        args.dir or dest_file, "yellow"))

    i = 0
    for config in configs:
        rel_path = config["group"] if config["group"] != DEFAULT_GROUP_NAME else ""
        if args.dir:
            try:
                os.makedirs(os.path.join(args.dir, rel_path))
            except OSError:
                pass
        i += 1
        sys.stdout.write("\033[K\rExporting: %s/%s   %s:%s" % (i, len(configs), config["group"], config["dataId"]))
        sys.stdout.flush()
        try:
            content = c.get(config["dataId"], config["group"], no_snapshot=True)
        except:
            print("Get content of %s:%s failed." % (config["group"] or DEFAULT_GROUP_NAME, config["dataId"]))
            sys.exit(1)

        if args.dir:
            _write_file(os.path.join(args.dir, rel_path, config["dataId"]), content)
        else:
            zip_file.writestr(os.path.join(rel_path, config["dataId"]), content.encode("utf8"))
    if zip_file:
        zip_file.close()
    print("")
    print("All dataIds exported.\n")


def import_to_server(args):
    e, ep, n, ns = _process_namespace(args)
    c = _get_client(e, ep, n, ns)

    if args.dir and not os.path.isdir(args.dir):
        print("%s does not exist." % _colored(args.dir, "red"))
        sys.exit(1)

    src_file = args.file or args.file or "%s-%s.zip" % (e, n)
    zip_file = None
    if not args.dir and not os.path.isfile(src_file):
        print("%s does not exist." % _colored(src_file, "red"))
        sys.exit(1)

    data_to_import = list()
    if args.dir:
        for f in os.listdir(args.dir):
            if f.startswith("."):
                continue
            if os.path.isfile(os.path.join(args.dir, f)):
                data_to_import.append((f, DEFAULT_GROUP_NAME))
            else:
                for ff in os.listdir(os.path.join(args.dir, f)):
                    if not ff.startswith(".") and os.path.isfile(os.path.join(args.dir, f, ff)):
                        data_to_import.append((ff, f))
    else:
        zip_file = zipfile.ZipFile(src_file, 'r', zipfile.ZIP_DEFLATED)
        for info in zip_file.infolist():
            sp = info.filename.split(os.path.sep)
            if len(sp) == 1:
                data_to_import.append((sp[0], DEFAULT_GROUP_NAME))
            elif len(sp) == 2 and sp[1]:
                data_to_import.append((sp[1], sp[0]))
            else:
                print("ignoring invalid path: %s" % info.filename)

    # process deleting
    if args.delete:
        # pick up candidates
        delete_list = list()
        configs = c.list_all()
        for i in configs:
            if (i["dataId"], i["group"]) not in data_to_import:
                delete_list.append(i)

        # deleting
        if delete_list:
            print("Following dataIds are not exist in %s:\n" % _colored(args.dir or src_file, "yellow"))
            for i in delete_list:
                print(" - %s:%s" % (i["group"], i["dataId"]))

            delete = True
            if not args.force:
                while True:
                    if sys.version_info[0] == 3:
                        choice = input("\nDeleting all dataIds above in ACM server? (y/n)")
                    else:
                        choice = raw_input("\nDeleting all dataIds above in ACM server? (y/n)")
                    if choice.lower() in ["y", "n"]:
                        delete = choice.lower() == "y"
                        break
                    print("Invalid choice, please input y or n.")
            if delete:
                for i in delete_list:
                    c.remove(i["dataId"], i["group"])
                print("Delete complete, continue to import...\n")

    print(_colored(len(data_to_import), "green") + " files will be imported to ACM server.\n")

    i = 0
    for data in data_to_import:
        i += 1
        sys.stdout.write("\033[K\rImporting: %s/%s   %s:%s" % (i, len(data_to_import), data[1], data[0]))
        sys.stdout.flush()
        if args.dir:
            f = os.path.join(args.dir, data[1], data[0]) if data[1] != DEFAULT_GROUP_NAME else os.path.join(args.dir,
                                                                                                            data[0])
            content = _read_file(f)
        else:
            name = os.path.join(data[1], data[0]) if data[1] != DEFAULT_GROUP_NAME else data[0]
            content = zip_file.read(name)

        try:
            c.publish(data[0], data[1], content)
        except:
            print("Publish %s/%s failed." % (data[1], data[0]))
            sys.exit(1)
    if zip_file:
        zip_file.close()
    print("")
    print("All files imported.\n")


def _get_client(e, ep, n, ns):
    c = ACMClient(endpoint=e, namespace=(None if n == "[default]" else n), ak=ns["ak"], sk=ns["sk"])
    if ep.get("kms_enabled"):
        c.set_options(kms_enabled=True, kms_ak=ns.get("kms_ak"), kms_secret=ns.get("kms_secret"),
                      region_id=ep.get("region_id"), key_id=ns.get("key_id"))
    if ep["tls"]:
        c.set_options(tls_enabled=True)
    return c


def arg_parse():
    parser = argparse.ArgumentParser(prog="acm",
                                     description="ACM command line tools for querying and exporting data.", )
    subparsers = parser.add_subparsers(help='sub-command help', title="Sub commands")

    # add
    parser_add = subparsers.add_parser("add", help="add a namespace",
                                       description='Add a namespace, '
                                                   'update if namespace is already exist.',
                                       epilog="Example: acm add acm.aliyun.com:ea61357b-d417-460c-92e4-032677dd8153 "
                                              "-s 'GLff***xcao=' -a 654b43******e9750 -n foo")
    parser_add.add_argument("namespace", default=None, help='use "endpoint:namespace_id" to locate a namespace, '
                                                            'if endpoint is missing, "acm.aliyun.com" act as default.')
    parser_add.add_argument("-a", dest="ak", default=None, help='AccessKey of this namespace.')
    parser_add.add_argument("-s", dest="sk", help='SecretKey of this namespace.')
    parser_add.add_argument("-n", dest="alias", help='alias of the namespace, ":" is not allowed in alias.')
    parser_add.add_argument("--tls", action="store_true", default=False, help="to use TLS connection.")
    parser_add.add_argument("--kms", action="store_true", default=False, help="to use Key Management Service (KMS).")
    parser_add.add_argument("-ka", dest="kms_ak", default=None,
                            help='AccessKey for KMS, use AccessKey by default, required if KMS is enabled.')
    parser_add.add_argument("-ks", dest="kms_secret", default=None,
                            help='SecretKey for KMS, use SecretKey by default, required if KMS is enabled.')
    parser_add.add_argument("-k", dest="key_id", default=None, help='Key ID of KMS, required if KMS is enabled.')
    parser_add.add_argument("-r", dest="region_id", default=None,
                            help='Region ID of Alibaba Cloud, required if KMS is enabled.')
    parser_add.set_defaults(func=add)

    # use
    parser_use = subparsers.add_parser("use", help="switch to a namespace",
                                       description="Switch to a namespace.",
                                       epilog="Example: acm use acm.aliyun.com:ea61357b-d417-460c-92e4-032677dd8153")
    parser_use.add_argument("namespace", help='"endpoint:namespace_id" or alias to use.')
    parser_use.set_defaults(func=use)

    # current
    parser_current = subparsers.add_parser("current", help="show current namespace",
                                           description="Show current namespace.")
    parser_current.set_defaults(func=current)

    # show
    parser_show = subparsers.add_parser("show", help="show all namespaces",
                                        description="Show all namespaces.")
    parser_show.set_defaults(func=show)

    # list
    parser_list = subparsers.add_parser("list", help="get list of dataIds", description="Get list of dataIds.")
    parser_list.add_argument("-g", dest="group", default=None, help='group of the dataId.')
    parser_list.add_argument("-p", dest="prefix", default=None, help='prefix of dataId.')
    parser_list.add_argument("-n", dest="namespace", default=None, help='"endpoint:namespace_id" or alias.')
    parser_list.set_defaults(func=list_conf)

    # pull
    parser_pull = subparsers.add_parser("pull", help="get one config content",
                                        description="Get one config content from ACM server.",
                                        epilog="Example: acm pull group/dataId > dest.txt")
    parser_pull.add_argument("data_id", help='the dataId to pull from, use group"/"dataId to specify group.')
    parser_pull.add_argument("-n", dest="namespace", default=None, help='"endpoint:namespace_id" or alias.')
    parser_pull.set_defaults(func=pull)

    # push
    parser_push = subparsers.add_parser("push", help="push one config",
                                        description="Push one config with the content of a local file or stdin.",
                                        epilog="Example: cat source.txt | acm push group/dataId")
    parser_push.add_argument("data_id", help='the dataId to store the content, use group"/"dataId to specify group.')
    parser_push.add_argument("-f", dest="file", default=None, help='the file to push, stdin can not be empty '
                                                                   'if file is not specified.')
    parser_push.add_argument("-n", dest="namespace", default=None, help='"endpoint:namespace_id" or alias.')
    parser_push.set_defaults(func=push)

    # export
    parser_export = subparsers.add_parser("export", help="export dataIds to local",
                                          description="Export dataIds of specified namespace to local dir or zip file.")
    parser_export.add_argument("-f", dest="file", default=None, help='zip file name, '
                                                                     'use "endpoint-namepspace_id.zip" as default.')
    parser_export.add_argument("-d", dest="dir", default=None, help='export destination dir, file is ignored '
                                                                    'if dir is specified.')
    parser_export.add_argument("-n", dest="namespace", default=None, help='"endpoint:namespace_id" or alias.')
    parser_export.add_argument("--delete", action="store_true", default=False,
                               help="[only for dir mode] "
                                    "delete the file not exist in ACM server (hidden files startswith . are igonred).")
    parser_export.add_argument("--force", action="store_true", default=False, help="[only for dir mode] "
                                                                                   "run and delete silently.")
    parser_export.set_defaults(func=export)

    # import
    parser_import = subparsers.add_parser("import", help="import local dir or zip file to ACM server",
                                          description="Import local dir or zip file to ACM server.")
    parser_import.add_argument("-f", dest="file", default=None, help='zip file name, '
                                                                     'use "endpoint-namepspace_id.zip" as default.')
    parser_import.add_argument("-d", dest="dir", default=None, help='import source dir, file is ignored '
                                                                    'if dir is specified.')
    parser_import.add_argument("-n", dest="namespace", default=None, help='"endpoint:namespace_id" or alias.')
    parser_import.add_argument("--delete", action="store_true", default=False,
                               help="delete the dataId not exist locally.")
    parser_import.add_argument("--force", action="store_true", default=False, help="run and delete silently.")
    parser_import.set_defaults(func=import_to_server)

    return parser.parse_args()


def main():
    args = arg_parse()
    try:
        args.func(args)
    except AttributeError:
        print("No sub command is specified, use -h for help.")

    """
    Usage:
        acm <command> [params]

    Global Commands:
        acm add {endpoint}:{namespace_id} [-a {ak} -s {sk}] [-n {alias}] # Add or update an namespace, not mandatory.
        acm use {endpoint}:{namespace}/{alias} # Choose a namespace for following commands (can be overwritten by -n).
        acm current # Print current endpoint and
        acm show # Print all endpoint and namespace info.

    Namespace Commands(use blank namespace if not specified):
        acm list [-g {group}] [-p {prefix}] # Get all dataIds matching group or prefix.

        acm pull {dataId}                   # Get a config content, default group is DEFAULT_GROUP.

        acm push {dataId} [-f {file}]       # Push one file or content from stdin to ACM

        acm export [-d {dir}] [-f {zip_file}] [--delete] [--force]  # Export dataIds as files.
            --delete:   If local file or directory can not match dataIds ACM, delete it.
            --force:    Overwrite or delete files without asking.

        acm import [-d {dir}] [-f {zip_file}] [--delete] [--force]  # Import files to ACM.
            --delete:   If dataId or group can not match local files, delete it.
            --force:    Overwrite or delete dataIds without asking.

    Examples:
        Configurations:
            |--------------|----------|------------------|
            |   Namespace  |   group  |      dataId      |
            +--------------+----------+------------------+
            |           dev|       ACM|   jdbc.properties|
            +--------------+----------+------------------+
            |           dev|       ACM|   application.yml|
            +--------------+----------+------------------+
            |           dev| VipServer|        nginx.conf|
            +--------------+----------+------------------+
            |       product|       ACM|   jdbc.properties|
            +--------------+----------+------------------+
            |       product|       ACM|   application.yml|
            +--------------+----------+------------------+
            |       product| VipServer|        nginx.conf|
            +--------------+----------+------------------+

        Get started:
            acm add acm.aliyun.com:'60*****0d38' -a '654b4*****50' -s 'GLf****ao=' -n dev
            acm add acm.aliyun.com:'7u*****d9eb' -a '37ab8*****2b' -s 'XMV****xc=' -n product

            acm use dev

        Find all configurations:
            acm list -g ACM

        Get from ACM:
            acm pull ACM/jdbc.properties >> ACM/jdbc.properties

        Modify it, then publish to ACM:
            cat ACM/jdbc.properties | acm push ACM/jdbc.properties

        Clone all configs from ACM:
            acm export -d dev_configs

        Local directories:
        dev_configs
          |
          |-- ACM
          |    |--jdbc.properties
          |    |--application.yml
          |
          |-- VipServer
               |--nginx.conf

        Transfer configs to an other namespace to make an product release
            acm import -d dev_configs -n product


        Tricks Use crontab and local VCS (git,svn) to make synchronization of configurations:
            acm use dev
            acm import -d {local_repo} --delete --force

    """


if __name__ == "__main__":
    main()
